@@ -1,4 +1,4 @@
-# frozen_string_literal: false
+# frozen_string_literal: true
 #
 # optparse.rb - command-line option analysis with the OptionParser class.
 #
@@ -125,6 +125,7 @@
 # For options that require an argument, option specification strings may include an
 # option name in all caps. If an option is used without the required argument,
 # an exception will be raised.
+#
 #   require 'optparse'
 #
 #   options = {}
@@ -137,9 +138,9 @@
 #
 # Used:
 #
-#   bash-3.2$ ruby optparse-test.rb -r
+#   $ ruby optparse-test.rb -r
 #   optparse-test.rb:9:in `<main>': missing argument: -r (OptionParser::MissingArgument)
-#   bash-3.2$ ruby optparse-test.rb -r my-library
+#   $ ruby optparse-test.rb -r my-library
 #   You required my-library!
 #
 # === Type Coercion
@@ -168,7 +169,7 @@
 # - Array -- Strings separated by ',' (e.g. 1,2,3)
 # - Regexp -- Regular expressions. Also includes options.
 #
-# We can also add our own coercions, which we will cover soon.
+# We can also add our own coercions, which we will cover below.
 #
 # ==== Using Built-in Conversions
 #
@@ -187,13 +188,12 @@
 #   end.parse!
 #
 # Used:
-#   bash-3.2$ ruby optparse-test.rb  -t nonsense
+#
+#   $ ruby optparse-test.rb  -t nonsense
 #   ... invalid argument: -t nonsense (OptionParser::InvalidArgument)
-#   from ... time.rb:5:in `block in <top (required)>'
-#   from optparse-test.rb:31:in `<main>'
-#   bash-3.2$ ruby optparse-test.rb  -t 10-11-12
+#   $ ruby optparse-test.rb  -t 10-11-12
 #   2010-11-12 00:00:00 -0500
-#   bash-3.2$ ruby optparse-test.rb  -t 9:30
+#   $ ruby optparse-test.rb  -t 9:30
 #   2014-08-13 09:30:00 -0400
 #
 # ==== Creating Custom Conversions
@@ -225,13 +225,39 @@
 #
 #   op.parse!
 #
-# output:
-#   bash-3.2$ ruby optparse-test.rb --user 1
+# Used:
+#
+#   $ ruby optparse-test.rb --user 1
 #   #<struct User id=1, name="Sam">
-#   bash-3.2$ ruby optparse-test.rb --user 2
+#   $ ruby optparse-test.rb --user 2
 #   #<struct User id=2, name="Gandalf">
-#   bash-3.2$ ruby optparse-test.rb --user 3
+#   $ ruby optparse-test.rb --user 3
 #   optparse-test.rb:15:in `block in find_user': No User Found for id 3 (RuntimeError)
+#
+# === Store options to a Hash
+#
+# The +into+ option of +order+, +parse+ and so on methods stores command line options into a Hash.
+#
+#   require 'optparse'
+#
+#   params = {}
+#   OptionParser.new do |opts|
+#     opts.on('-a')
+#     opts.on('-b NUM', Integer)
+#     opts.on('-v', '--verbose')
+#   end.parse!(into: params)
+#
+#   p params
+#
+# Used:
+#
+#   $ ruby optparse-test.rb -a
+#   {:a=>true}
+#   $ ruby optparse-test.rb -a -v
+#   {:a=>true, :verbose=>true}
+#   $ ruby optparse-test.rb -a -b 100
+#   {:a=>true, :b=>100}
+#
 # === Complete example
 #
 # The following example is a complete Ruby program.  You can run it and see the
@@ -413,7 +439,7 @@ class OptionParser
       candidates = []
       block.call do |k, *v|
         (if Regexp === k
-           kn = "".freeze
+           kn = ""
            k === key
          else
            kn = defined?(k.id2name) ? k.id2name : k
@@ -508,8 +534,9 @@ class OptionParser
 
     def initialize(pattern = nil, conv = nil,
                    short = nil, long = nil, arg = nil,
-                   desc = ([] if short or long), block = Proc.new)
+                   desc = ([] if short or long), block = nil, &_block)
       raise if Array === pattern
+      block ||= _block
       @pattern, @conv, @short, @long, @arg, @desc, @block =
         pattern, conv, short, long, arg, desc, block
     end
@@ -565,7 +592,7 @@ class OptionParser
     #            +max+ columns.
     # +indent+:: Prefix string indents all summarized lines.
     #
-    def summarize(sdone = [], ldone = [], width = 1, max = width - 1, indent = "")
+    def summarize(sdone = {}, ldone = {}, width = 1, max = width - 1, indent = "")
       sopts, lopts = [], [], nil
       @short.each {|s| sdone.fetch(s) {sopts << s}; sdone[s] = true} if @short
       @long.each {|s| ldone.fetch(s) {lopts << s}; ldone[s] = true} if @long
@@ -577,7 +604,7 @@ class OptionParser
       while s = lopts.shift
         l = left[-1].length + s.length
         l += arg.length if left.size == 1 && arg
-        l < max or sopts.empty? or left << ''
+        l < max or sopts.empty? or left << +''
         left[-1] << (left[-1].empty? ? ' ' * 4 : ', ') << s
       end
 
@@ -628,7 +655,7 @@ class OptionParser
       return if sopts.empty? and lopts.empty? # completely hidden
 
       (sopts+lopts).each do |opt|
-        # "(-x -c -r)-l[left justify]" \
+        # "(-x -c -r)-l[left justify]"
         if /^--\[no-\](.+)$/ =~ opt
           o = $1
           yield("--#{o}", desc.join(""))
@@ -839,6 +866,10 @@ class OptionParser
       __send__(id).complete(opt, icase, *pat, &block)
     end
 
+    def get_candidates(id)
+      yield __send__(id).keys
+    end
+
     #
     # Iterates over each option, passing the option to the +block+.
     #
@@ -1136,7 +1167,7 @@ XXX
   #
   def banner
     unless @banner
-      @banner = "Usage: #{program_name} [options]"
+      @banner = +"Usage: #{program_name} [options]"
       visit(:add_banner, @banner)
     end
     @banner
@@ -1165,14 +1196,14 @@ XXX
   # Version
   #
   def version
-    @version || (defined?(::Version) && ::Version)
+    (defined?(@version) && @version) || (defined?(::Version) && ::Version)
   end
 
   #
   # Release code
   #
   def release
-    @release || (defined?(::Release) && ::Release) || (defined?(::RELEASE) && ::RELEASE)
+    (defined?(@release) && @release) || (defined?(::Release) && ::Release) || (defined?(::RELEASE) && ::RELEASE)
   end
 
   #
@@ -1180,7 +1211,7 @@ XXX
   #
   def ver
     if v = version
-      str = "#{program_name} #{[v].join('.')}"
+      str = +"#{program_name} #{[v].join('.')}"
       str << " (#{v})" if v = release
       str
     end
@@ -1237,7 +1268,8 @@ XXX
   # +indent+:: Indentation, defaults to @summary_indent.
   #
   def summarize(to = [], width = @summary_width, max = width - 1, indent = @summary_indent, &blk)
-    blk ||= proc {|l| to << (l.index($/, -1) ? l : l + $/)}
+    nl = "\n"
+    blk ||= proc {|l| to << (l.index(nl, -1) ? l : l + nl)}
     visit(:summarize, {}, {}, width, max, indent, &blk)
     to
   end
@@ -1324,6 +1356,8 @@ XXX
   # [Description:]
   #   Description string for the option.
   #     "Run verbosely"
+  #   If you give multiple description strings, each string will be printed
+  #   line by line.
   #
   # [Handler:]
   #   Handler for the parsed argument value. Either give a block or pass a
@@ -1514,7 +1548,10 @@ XXX
 
   #
   # Parses command line arguments +argv+ in order. When a block is given,
-  # each non-option argument is yielded.
+  # each non-option argument is yielded. When optional +into+ keyword
+  # argument is provided, the parsed option values are stored there via
+  # <code>[]=</code> method (so it can be Hash, or OpenStruct, or other
+  # similar object).
   #
   # Returns the rest of +argv+ left unparsed.
   #
@@ -1565,7 +1602,7 @@ XXX
               begin
                 sw, = complete(:short, opt)
                 # short option matched.
-                val = arg.sub(/\A-/, '')
+                val = arg.delete_prefix('-')
                 has_arg = true
               rescue InvalidOption
                 # if no short options match, try completion with long
@@ -1610,7 +1647,10 @@ XXX
 
   #
   # Parses command line arguments +argv+ in permutation mode and returns
-  # list of non-option arguments.
+  # list of non-option arguments. When optional +into+ keyword
+  # argument is provided, the parsed option values are stored there via
+  # <code>[]=</code> method (so it can be Hash, or OpenStruct, or other
+  # similar object).
   #
   def permute(*argv, into: nil)
     argv = argv[0].dup if argv.size == 1 and Array === argv[0]
@@ -1631,6 +1671,9 @@ XXX
   #
   # Parses command line arguments +argv+ in order when environment variable
   # POSIXLY_CORRECT is set, and in permutation mode otherwise.
+  # When optional +into+ keyword argument is provided, the parsed option
+  # values are stored there via <code>[]=</code> method (so it can be Hash,
+  # or OpenStruct, or other similar object).
   #
   def parse(*argv, into: nil)
     argv = argv[0].dup if argv.size == 1 and Array === argv[0]
@@ -1653,11 +1696,11 @@ XXX
   # Wrapper method for getopts.rb.
   #
   #   params = ARGV.getopts("ab:", "foo", "bar:", "zot:Z;zot option")
-  #   # params[:a] = true   # -a
-  #   # params[:b] = "1"    # -b1
-  #   # params[:foo] = "1"  # --foo
-  #   # params[:bar] = "x"  # --bar x
-  #   # params[:zot] = "z"  # --zot Z
+  #   # params["a"] = true   # -a
+  #   # params["b"] = "1"    # -b1
+  #   # params["foo"] = "1"  # --foo
+  #   # params["bar"] = "x"  # --bar x
+  #   # params["zot"] = "z"  # --zot Z
   #
   def getopts(*args)
     argv = Array === args.first ? args.shift : default_argv
@@ -1734,26 +1777,41 @@ XXX
     if pat.empty?
       search(typ, opt) {|sw| return [sw, opt]} # exact match or...
     end
-    raise AmbiguousOption, catch(:ambiguous) {
+    ambiguous = catch(:ambiguous) {
       visit(:complete, typ, opt, icase, *pat) {|o, *sw| return sw}
-      raise InvalidOption, opt
     }
+    exc = ambiguous ? AmbiguousOption : InvalidOption
+    raise exc.new(opt, additional: self.method(:additional_message).curry[typ])
   end
   private :complete
 
+  #
+  # Returns additional info.
+  #
+  def additional_message(typ, opt)
+    return unless typ and opt and defined?(DidYouMean::SpellChecker)
+    all_candidates = []
+    visit(:get_candidates, typ) do |candidates|
+      all_candidates.concat(candidates)
+    end
+    all_candidates.select! {|cand| cand.is_a?(String) }
+    checker = DidYouMean::SpellChecker.new(dictionary: all_candidates)
+    DidYouMean.formatter.message_for(all_candidates & checker.correct(opt))
+  end
+
   def candidate(word)
     list = []
     case word
+    when '-'
+      long = short = true
     when /\A--/
       word, arg = word.split(/=/, 2)
       argpat = Completion.regexp(arg, false) if arg and !arg.empty?
       long = true
-    when /\A-(!-)/
-      short = true
     when /\A-/
-      long = short = true
+      short = true
     end
-    pat = Completion.regexp(word, true)
+    pat = Completion.regexp(word, long)
     visit(:each_option) do |opt|
       next unless Switch === opt
       opts = (long ? opt.long : []) + (short ? opt.short : [])
@@ -1776,13 +1834,26 @@ XXX
   # is not present. Returns whether successfully loaded.
   #
   # +filename+ defaults to basename of the program without suffix in a
-  # directory ~/.options.
+  # directory ~/.options, then the basename with '.options' suffix
+  # under XDG and Haiku standard places.
   #
   def load(filename = nil)
-    begin
-      filename ||= File.expand_path(File.basename($0, '.*'), '~/.options')
-    rescue
-      return false
+    unless filename
+      basename = File.basename($0, '.*')
+      return true if load(File.expand_path(basename, '~/.options')) rescue nil
+      basename << ".options"
+      return [
+        # XDG
+        ENV['XDG_CONFIG_HOME'],
+        '~/.config',
+        *ENV['XDG_CONFIG_DIRS']&.split(File::PATH_SEPARATOR),
+
+        # Haiku
+        '~/config/settings',
+      ].any? {|dir|
+        next if !dir or dir.empty?
+        load(File.expand_path(basename, dir)) rescue nil
+      }
     end
     begin
       parse(*IO.readlines(filename).each {|s| s.chomp!})
@@ -1842,7 +1913,7 @@ XXX
   #
   # Float number format, and converts to Float.
   #
-  float = "(?:#{decimal}(?:\\.(?:#{decimal})?)?|\\.#{decimal})(?:E[-+]?#{decimal})?"
+  float = "(?:#{decimal}(?=(.)?)(?:\\.(?:#{decimal})?)?|\\.#{decimal})(?:E[-+]?#{decimal})?"
   floatpat = %r"\A[-+]?#{float}\z"io
   accept(Float, floatpat) {|s,| s.to_f if s}
 
@@ -1851,11 +1922,13 @@ XXX
   # for float format, and Rational for rational format.
   #
   real = "[-+]?(?:#{octal}|#{float})"
-  accept(Numeric, /\A(#{real})(?:\/(#{real}))?\z/io) {|s, d, n|
+  accept(Numeric, /\A(#{real})(?:\/(#{real}))?\z/io) {|s, d, f, n,|
     if n
       Rational(d, n)
-    elsif s
-      eval(s)
+    elsif f
+      Float(s)
+    else
+      Integer(s)
     end
   }
 
@@ -1865,7 +1938,7 @@ XXX
   DecimalInteger = /\A[-+]?#{decimal}\z/io
   accept(DecimalInteger, DecimalInteger) {|s,|
     begin
-      Integer(s)
+      Integer(s, 10)
     rescue ArgumentError
       raise OptionParser::InvalidArgument, s
     end if s
@@ -1889,10 +1962,14 @@ XXX
   # integer format, Float for float format.
   #
   DecimalNumeric = floatpat     # decimal integer is allowed as float also.
-  accept(DecimalNumeric, floatpat) {|s,|
+  accept(DecimalNumeric, floatpat) {|s, f|
     begin
-      eval(s)
-    rescue SyntaxError
+      if f
+        Float(s)
+      else
+        Integer(s)
+      end
+    rescue ArgumentError
       raise OptionParser::InvalidArgument, s
     end if s
   }
@@ -1946,15 +2023,18 @@ XXX
   #
   class ParseError < RuntimeError
     # Reason which caused the error.
-    Reason = 'parse error'.freeze
+    Reason = 'parse error'
 
-    def initialize(*args)
+    def initialize(*args, additional: nil)
+      @additional = additional
+      @arg0, = args
       @args = args
       @reason = nil
     end
 
     attr_reader :args
     attr_writer :reason
+    attr_accessor :additional
 
     #
     # Pushes back erred argument(s) to +argv+.
@@ -1999,7 +2079,7 @@ XXX
     # Default stringizing method to emit standard error message.
     #
     def message
-      reason + ': ' + args.join(' ')
+      "#{reason}: #{args.join(' ')}#{additional[@arg0] if additional}"
     end
 
     alias to_s message
@@ -2009,42 +2089,42 @@ XXX
   # Raises when ambiguously completable string is encountered.
   #
   class AmbiguousOption < ParseError
-    const_set(:Reason, 'ambiguous option'.freeze)
+    const_set(:Reason, 'ambiguous option')
   end
 
   #
   # Raises when there is an argument for a switch which takes no argument.
   #
   class NeedlessArgument < ParseError
-    const_set(:Reason, 'needless argument'.freeze)
+    const_set(:Reason, 'needless argument')
   end
 
   #
   # Raises when a switch with mandatory argument has no argument.
   #
   class MissingArgument < ParseError
-    const_set(:Reason, 'missing argument'.freeze)
+    const_set(:Reason, 'missing argument')
   end
 
   #
   # Raises when switch is undefined.
   #
   class InvalidOption < ParseError
-    const_set(:Reason, 'invalid option'.freeze)
+    const_set(:Reason, 'invalid option')
   end
 
   #
   # Raises when the given argument does not match required format.
   #
   class InvalidArgument < ParseError
-    const_set(:Reason, 'invalid argument'.freeze)
+    const_set(:Reason, 'invalid argument')
   end
 
   #
   # Raises when the given argument word can't be completed uniquely.
   #
   class AmbiguousArgument < InvalidArgument
-    const_set(:Reason, 'ambiguous argument'.freeze)
+    const_set(:Reason, 'ambiguous argument')
   end
 
   #
